import os
from multiprocessing import Event, Process

import keyring
import requests
import segment.analytics as analytics
import typer

from mirageml.constants import ANALYTICS_WRITE_KEY, SERVICE_ID, USER_CHECK_ENDPOINT, supabase

analytics.write_key = ANALYTICS_WRITE_KEY


def login_process(e, already_exists, email, password):
    try:
        if already_exists:
            response = supabase.auth.sign_in_with_password({"email": email, "password": password})
        else:
            response = supabase.auth.sign_up({"email": email, "password": password})

        user = response.user
        session = response.session

        user_id = user.id
        email = user.email
        access_token = session.access_token
        refresh_token = session.refresh_token
        expires_at = session.expires_at

        keyring.set_password(SERVICE_ID, "user_id", user_id)
        keyring.set_password(SERVICE_ID, "email", email)
        keyring.set_password(SERVICE_ID, "access_token", access_token)
        keyring.set_password(SERVICE_ID, "refresh_token", refresh_token)
        keyring.set_password(SERVICE_ID, "expires_at", str(expires_at))
        analytics.identify(
            user_id,
            {
                "email": email,
            },
        )
        e.set()
        return
    except Exception as e:
        print(e)
        typer.secho("Error logging in. Please try again", fg=typer.colors.BRIGHT_RED, bold=True)
        os._exit(1)


def login():
    try:
        already_exists = False
        email = typer.prompt("Email")
        if email.strip() != "":
            already_exists = requests.post(USER_CHECK_ENDPOINT, json={"email": email}).json()["exists"]

        if not already_exists:
            print("Looks like you don't have an account yet. Let's create one!")
        password = typer.prompt(
            "Password", hide_input=True, confirmation_prompt="Confirm Password" if not already_exists else None
        )

        # Use multiprocessing to end the process if it takes too long
        e = Event()
        p = Process(target=login_process, args=(e, already_exists, email, password))
        p.start()
        e.wait()
        p.terminate()

        typer.secho("Successfully logged in", fg=typer.colors.BRIGHT_GREEN, bold=True)
        return
    except Exception as e:
        print(e)
        typer.secho("Error logging in. Please try again", fg=typer.colors.BRIGHT_RED, bold=True)
        os._exit(1)



from .utils.vectordb import delete_local_qdrant_db, delete_remote_qdrant_db


def delete_source(names: list[str]):
    import typer

    from .config import load_config

    config = load_config()
    sources = config["local"] + config["remote"]

    for name in names:
        while True:
            if name not in sources:
                typer.secho(
                    f"Source: {name} does not exist. Choose one of the following or exit:",
                    fg=typer.colors.BRIGHT_RED,
                )
                print("\n".join(sources))
                name = typer.prompt("Enter source name", default="exit", show_default=False)
                if not name or name == "exit":
                    return
            else:
                break

        typer.secho(f"Deleting Source: {name}...", fg=typer.colors.BRIGHT_RED)
        delete_local_qdrant_db(name)
        delete_remote_qdrant_db(name)
        print(f"Deleted Source: {name}")



import keyring
import typer

from ..constants import SERVICE_ID


def profile():
    email = keyring.get_password(SERVICE_ID, "email")
    typer.secho("Current Profile", fg=typer.colors.BRIGHT_GREEN, bold=True)
    typer.echo(f"Email: {email}")



import typer

from .config import append_var_config


def fix_name(name):
    if name.startswith("http"):
        name = "_".join(name.split("/")[1:])

    name = name.replace(" ", "_")
    name = name.replace("/", "_")
    if name.startswith("_") or name.startswith("/"):
        name = name[1:]
    return name


def add_system_prompt(args):
    name = args["name"]
    if not name:
        name = typer.prompt("Name")
    name = fix_name(name)
    prompt = typer.prompt("Prompt")
    sp = {"name": name, "prompt": prompt}
    append_var_config("system_prompts", sp)
    typer.secho(f"Added system prompt: {name}", fg=typer.colors.BRIGHT_GREEN, bold=True)



import os
import sys
from io import StringIO

import keyring
import requests

from ...constants import (
    LLM_GPT_ENDPOINT,
    SERVICE_ID,
    get_headers,
)
from ..config import load_config

PACKAGE_DIR = os.path.dirname(__file__)
os.environ["TRANSFORMERS_CACHE"] = os.path.join(PACKAGE_DIR, "models")
os.environ["NUMEXPR_MAX_THREADS"] = "8"


def _chunk_data(data, metadata):
    from langchain.text_splitter import RecursiveCharacterTextSplitter

    text_splitter = RecursiveCharacterTextSplitter(chunk_size=2048, chunk_overlap=80)
    docs = text_splitter.create_documents([data])
    chunks = [x.page_content for x in docs]
    meta = [{"data": curr_chunk, "source": metadata["source"]} for curr_chunk in chunks]
    vector_data = local_get_embedding(chunks)

    return chunks, meta, vector_data


def local_get_embedding(text_list, embedding_model_id="BAAI/bge-base-en-v1.5"):
    model_dir = os.path.join(PACKAGE_DIR, "models", embedding_model_id)
    if not os.path.exists(model_dir):
        os.makedirs(model_dir, exist_ok=True)
        print("Downloading model to:", model_dir)
        print("This will take a few minutes and only happen once!")

    # TODO: Suppress stdout/stderr
    original_stdout, original_stderr = sys.stdout, sys.stderr
    sys.stdout, sys.stderr = StringIO(), StringIO()

    # # Restore stdout/stderr
    from fastembed.embedding import FlagEmbedding as Embedding

    embedding_model = Embedding(model_name=embedding_model_id, max_length=512, cache_dir=model_dir)
    embeddings = list(embedding_model.embed(text_list))
    sys.stdout, sys.stderr = original_stdout, original_stderr

    # Convert the embeddings to a list
    embeddings = [x.tolist() for x in embeddings]  # size = 768
    return embeddings


def local_llm_call(messages, llm_model_id="TheBloke/Llama-2-7b-Chat-GGUF", stream=False, live=None):
    from ctransformers import AutoModelForCausalLM
    from rich.box import HORIZONTALS
    from rich.panel import Panel

    os.environ["TOKENIZERS_PARALLELISM"] = "false"

    model_dir = os.path.join(PACKAGE_DIR, "models", llm_model_id)
    if not os.path.exists(model_dir):
        os.makedirs(model_dir, exist_ok=True)
        print("Downloading model to:", model_dir)
        if live:
            live.update(
                Panel(
                    "This will take a few minutes and only happen once!",
                    title="[bold blue]Assistant[/bold blue]",
                    box=HORIZONTALS,
                    border_style="blue",
                )
            )

    # Suppress stdout/stderr
    original_stdout, original_stderr = sys.stdout, sys.stderr
    sys.stdout, sys.stderr = StringIO(), StringIO()

    llm = AutoModelForCausalLM.from_pretrained(
        "TheBloke/Llama-2-7b-Chat-GGUF",
        model_file="llama-2-7b-chat.Q4_K_M.gguf",
        model_type="llama",
    )

    # Restore stdout/stderr
    sys.stdout, sys.stderr = original_stdout, original_stderr

    ## Workaround for now, max context length is 512
    formatted_messages = "\n".join([x["content"] for x in messages])[-1000:]

    if stream:
        return llm(formatted_messages, stream=True)
    else:
        return llm(formatted_messages)


def get_embedding(text_list, model="BAAI/bge-small-en-v1.5"):
    raise NotImplementedError


def llm_call(messages, model="gpt-3.5-turbo", stream=False, local=False, live=None):
    if local:
        return local_llm_call(messages, stream=stream, live=live)
    config = load_config()
    openai_key = config["openai_key"] if len(config["openai_key"]) > 0 else None
    user_id = keyring.get_password(SERVICE_ID, "user_id")
    json_data = {"user_id": user_id, "model": model, "messages": messages, "stream": stream}
    if openai_key:
        json_data["openai_key"] = openai_key
    return requests.post(LLM_GPT_ENDPOINT, json=json_data, headers=get_headers(), stream=stream)



# Copyright Mirage ML Inc 2023
"""Specifies the `mirageml.__version__` number for the client package."""

from ._version_generated import build_number

# As long as we're on 0.*, all versions are published automatically
major_number = 1

# Bump this manually on any major changes
minor_number = 1

# TODO set the patch number (the 3rd field) to the job run number in GitHub
__version__ = f"{major_number}.{minor_number}.{build_number}"



import keyring
import requests
import typer

from ..constants import FINETUNE_CREATE_ENDPOINT, SERVICE_ID, get_headers


def fix_name(name):
    if name.startswith("http"):
        name = "-".join(name.split("/")[1:])

    name = name.replace(" ", "-")
    name = name.replace("/", "-")
    if name.startswith("_") or name.startswith("/"):
        name = name[1:]
    return name


def add_model(model_name, links):
    from .list_models import get_models

    model_name = fix_name(model_name)

    if model_name in get_models():
        typer.secho(f"Model {model_name} already exists", fg=typer.colors.RED, bold=True)
        return

    user_id = keyring.get_password(SERVICE_ID, "user_id")
    json_data = {"user_id": user_id, "finetune_model_name": model_name, "links": links}

    requests.post(FINETUNE_CREATE_ENDPOINT, json=json_data, headers=get_headers())
    typer.secho(
        f"Creating Finetuned Model: {model_name}. You will receive an email once its ready!",
        fg=typer.colors.GREEN,
        bold=True,
    )
    return True



import os

import typer
from rich import print
from rich.box import HORIZONTALS
from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from rich.panel import Panel

from .add_source import add_local_source
from .config import load_config
from .list_sources import get_sources
from .rag import rag_chat
from .utils.codeblocks import (
    add_indices_to_code_blocks,
    copy_code_to_clipboard,
    extract_code_from_markdown,
)
from .utils.custom_inputs import multiline_input
from .utils.llm import llm_call
from .utils.type_check import is_convertable_to_int

console = Console()
config = load_config()


def chat(files: list[str] = [], urls: list[str] = [], sources: list[str] = [], sp: str = ""):
    # Beginning of the chat sequence
    transient_sources = []
    if files or urls or sources:
        index_local = False
        if "local" in sources:
            index_local = True
            sources.remove("local")

        for file in files:
            if os.path.isdir(file):
                sources.append(add_local_source(file))
                files.remove(file)

        local_sources, remote_sources = get_sources()
        all_sources = list(set(local_sources + remote_sources))
        for source in sources:
            if source not in all_sources:
                typer.secho(
                    f"Source: {source} does not exist. Please add it with `mirageml add source`",
                    fg=typer.colors.RED,
                    bold=True,
                )
                sources.remove(source)

        if index_local:
            sources.append(add_local_source("local"))

        if files or urls:
            from .utils.local_source import crawl_files
            from .utils.web_source import extract_from_url

            with Live(
                Panel(
                    "Scraping Files and Urls",
                    title="[bold green]Assistant[/bold green]",
                    border_style="green",
                ),
                console=console,
                transient=True,
            ) as live:
                for source, extractor in zip(files + urls, [crawl_files] * len(files) + [extract_from_url] * len(urls)):
                    live.update(
                        Panel(
                            f"Extracting data from {source}",
                            title="[bold green]Assistant[/bold green]",
                            border_style="green",
                        )
                    )
                    data, metadata = extractor(source)
                    if data:
                        transient_sources.append((data, metadata))

    chat_history = [{"role": "system", "content": "You are a helpful assistant."}]
    while True:
        ai_response = ""
        code_blocks = []
        if sources or transient_sources:
            chat_history, ai_response = rag_chat(sources, transient_sources)

        set_sp = False
        if sp:
            new_system_prompts = list(filter(lambda x: x["name"] == sp, config["system_prompts"]))
            if len(new_system_prompts) == 0:
                typer.secho("Please enter a valid system prompt name", fg=typer.colors.RED, bold=True)
            else:
                set_sp = True
                chat_history[0]["content"] = new_system_prompts[0]["prompt"]

        while True:
            # Loop for follow-up questions
            try:
                if ai_response:
                    chat_history.append({"role": "assistant", "content": ai_response})
                    indexed_ai_response = add_indices_to_code_blocks(ai_response)
                    console.print(
                        Panel(
                            Markdown(indexed_ai_response),
                            title="[bold blue]Assistant[/bold blue]",
                            box=HORIZONTALS,
                            border_style="blue",
                        )
                    )
                    code_blocks = extract_code_from_markdown(ai_response)
                    if len(code_blocks) > 0:
                        user_input = multiline_input(
                            "Enter the number(s) of the code blocks you want to copy, separated by commas (e.g. '1,3,4') or ask a follow-up. Type reset to search again. Ctrl+C to interrupt"
                        )
                    else:
                        user_input = multiline_input("Ask a follow-up. Type reset to search again. Ctrl+C to interrupt")
                else:
                    if len(chat_history) == 1:
                        base_message = "Chat with Mirage"
                        if set_sp:
                            base_message += f" using system prompt: {sp}."
                        user_input = multiline_input(base_message)
                    elif len(code_blocks) > 0:
                        user_input = multiline_input(
                            "Enter the number(s) of the code blocks you want to copy, separated by commas (e.g. '1,3,4') or ask a follow-up. Type reset to search again. Ctrl+C to interrupt"
                        )
                    else:
                        user_input = multiline_input("Ask a follow-up. Type reset to search again. Ctrl+C to interrupt")

                if user_input.lower().strip() == "exit":
                    typer.secho("Ending chat. Goodbye!", fg=typer.colors.BRIGHT_GREEN, bold=True)
                    return
                elif user_input.lower().strip() == "/help" or user_input.lower().strip() == "help":
                    typer.secho("Use this chat to talk to mirage", fg=typer.colors.BRIGHT_GREEN, bold=True)
                    typer.echo("'exit' to end the chat.")
                    typer.echo("'reset' to start over.")
                    typer.echo("'/help' to see this message again.")
                    typer.echo(
                        "'/copy {index}' to copy a code block to your clipboard. (Note: code block indicies start at 1)"
                    )
                    typer.echo("'/sp set {system_prompt_name}' to set the system prompt.")
                    continue
                elif is_convertable_to_int(user_input.lower().strip()) and len(code_blocks) >= int(
                    user_input.lower().strip()
                ):
                    copy_code_to_clipboard(code_blocks, [int(user_input.lower().strip())])
                    ai_response = ""
                    continue
                elif user_input.lower().strip().startswith("/copy"):
                    user_input_split = user_input.split(" ")
                    if len(code_blocks) == 0:
                        typer.secho("No code blocks to copy", fg=typer.colors.RED, bold=True)
                    elif len(user_input_split) == 1:
                        typer.secho("Please enter a code block index", fg=typer.colors.RED, bold=True)
                    elif len(user_input_split) == 2:
                        try:
                            selected_indicies = list(map(int, user_input_split[1].split(",")))
                            copy_code_to_clipboard(code_blocks, selected_indicies)
                            typer.secho("Copied code to clipboard", fg=typer.colors.BRIGHT_GREEN, bold=True)
                            ai_response = ""
                        except Exception:
                            typer.secho(
                                "Could not copy. Please enter a valid code block index", fg=typer.colors.RED, bold=True
                            )
                    continue
                elif user_input.lower().strip().startswith("/sp"):
                    user_input_split = user_input.split(" ")
                    if len(user_input_split) == 1:
                        from .list_system_prompts import list_system_prompts

                        print(f"[bold green]Current System Prompt: [/green bold]{chat_history[0]['content']}")
                        list_system_prompts()
                    elif user_input_split[1] == "set" and len(user_input_split) == 2:
                        typer.secho("Please enter a system prompt name", fg=typer.colors.RED, bold=True)
                    elif user_input_split[1] == "set" and len(user_input_split == 3):
                        system_prompt_name = user_input_split[2]
                        new_system_prompts = list(
                            filter(lambda x: x["name"] == system_prompt_name, config["system_prompts"])
                        )
                        if len(new_system_prompts) == 0:
                            typer.secho("Please enter a valid system prompt name", fg=typer.colors.RED, bold=True)
                        else:
                            chat_history[0]["content"] = new_system_prompts[0]["prompt"]
                            typer.secho(
                                f"System prompt set to {system_prompt_name}", fg=typer.colors.BRIGHT_GREEN, bold=True
                            )
                    continue
                elif (
                    user_input.lower().strip() == "reset"
                    or user_input.lower().strip() == "restart"
                    or user_input.lower().strip() == "start over"
                    or user_input.lower().strip() == "clear"
                ):
                    break

            except KeyboardInterrupt:
                typer.secho("Ending chat. Goodbye!", fg=typer.colors.BRIGHT_GREEN, bold=True)
                return

            chat_history.append({"role": "user", "content": user_input})

            try:
                with Live(
                    Panel(
                        "Assistant is thinking...",
                        title="[bold blue]Assistant[/bold blue]",
                        box=HORIZONTALS,
                        border_style="blue",
                    ),
                    console=console,
                    transient=True,
                    auto_refresh=True,
                    refresh_per_second=8,
                ) as live:
                    response = llm_call(
                        chat_history,
                        model=config["model"],
                        stream=True,
                        local=config["local_mode"],
                    )

                    ai_response = ""
                    if config["local_mode"]:
                        for chunk in response:
                            ai_response += chunk
                            live.update(
                                Panel(
                                    Markdown(ai_response),
                                    title="[bold blue]Assistant[/bold blue]",
                                    box=HORIZONTALS,
                                    border_style="blue",
                                )
                            )
                    else:
                        for chunk in response.iter_content(chunk_size=512):
                            if chunk:
                                decoded_chunk = chunk.decode("utf-8")
                                ai_response += decoded_chunk
                                live.update(
                                    Panel(
                                        Markdown(ai_response),
                                        title="[bold blue]Assistant[/bold blue]",
                                        box=HORIZONTALS,
                                        border_style="blue",
                                    )
                                )
            except KeyboardInterrupt:
                pass
